import sys
import re

#object with the data read from the user, all in string mode

class RawData(object):
    def __init__(self, rawLogic, operation, dataOperation):
        #just stores all the input from the user in an practical object 
		self.logic= rawLogic
		#self.modal = modal
		self.operation = operation
		self.dataOperation = dataOperation


#object to store the processed data (the ordinal sum corresponding to the logic, the formulas adapted to the language, the constants map, the modal skeleton...
class Data(object):
    ###PUBLIC CONSTRUCTOR
    def __init__(self, rawData):
        self.operation = rawData.operation
        #self.modal = rawData.modal
        self._treatLogic(rawData) #self.ordinalSum + self.variables
        self.spEquations = []
        self._treatConstants(rawData) #self.dataOperation + self.constantsMap + self.constantsInProducts
        if self.operation == "s":
            for equation in self.dataOperation:
                self._splitEquation(equation)  #self.spEquations
        self._buildSkeleton() #self.skeleton, #self.nworlds
        
        #print self.ordinalSum
        #print self.skeleton
    
    

    #INTERNAL FUNCTIONS FOR THE TREATMENT OF THE LOGIC
    
    #processes the non-bl logics
    #bl or
    #[0,t1]l1+[t1,t2]l2+ ...+[tn-1,1]ln
    def _treatNoBLlogic(self, rawLogic):
        ordinalSum = []
        rcomponents = rawLogic.split('+')
        for c in rcomponents:
            pars = re.search('(\(.+,.+\))(.+)',c)
            logic = pars.group(2)
            rintervalc = pars.group(1) #interval in string
            intervalc = map(float, rintervalc[1:len(rintervalc)-1].split(','))#interval as pair of reals
            if logic == 'l' or logic == 'g' or logic == 'p':
                ordinalSum.append([intervalc, logic, 0])
            else:
				lukfin = re.search('(?<=l)([0-9]+)', logic)
				godfin = re.search('(?<=g)([0-9]+)', logic)
				if lukfin:
					ordinalSum.append([intervalc, "l", int(lukfin.group(0))])
				else:
					ordinalSum.append([intervalc, "g", int(godfin.group(0))])
        return ordinalSum
        
    #processes the input of the logic (either bl or in th format defined) and stores the results: ordinalSum
    # generates the ordinal sum: a list of triplas with the interval, the logic and the number of values (0 if infinite) 
    #and the list of variables present in the formulas/equations    
    def _treatLogic(self, rawData):
        modals = []
        allVars = []
        for form in rawData.dataOperation:
            allVars.extend(re.findall('(x[0-9]+)', form))
            modals.extend(re.findall('(box)|(diamond)', form))
        self.variables = list(set(allVars))
        totalVars = len(self.variables)
        
        if modals == []:
            self.modal = 0
        else:
            self.modal = 1
        
        if rawData.logic == "bl":
            self.ordinalSum = []
            #totalVars + 1 luk. fragments
            for i in range(totalVars+1):
                self.ordinalSum.append([[float(i)/(totalVars+1), float(i+1)/(totalVars+1)], "l", 0])
        else :
            self.ordinalSum = self._treatNoBLlogic(rawData.logic)
        return
            
    #FUNCTIONS FOR THE MANAGEMENT OF THE CONSTANTS
     


    def _vinc(self, component, value):
        #remember the shape of the component is c[0] = [a,b], c[1] = type, c[2] = nvalues
        #this does not control the finite-valued components, it is oriented towards the product ones
        return component[0][0] < value and value < component[0][1]
    
    #the constants will be substituted, by symbols, and its dictionary is added to the structure
    #the constants must be in float format (f.i, 0.42) or be equal to T or F
    def _treatConstants(self, rawData):
        
        #dataOperation = rawData.dataOperation
        
        #allVars = [re.findall('(x[0-9]+)', inputs) for inputs in dataOperation]
        #variables = list(set(allVars))
    	#totalVars = len(variables)
    	
    	#extremeConstants = map(int, re.findall(' (0)| (1)', inputs) for inputs in dataOperation])
        #for inp in dataOperation:
        #    inp.replace(" 1", " T")
        #    inp.replace(" 0", " F")
      
        tconstants = []
        for form in rawData.dataOperation:
            tconstants.extend(re.findall('(0\.[0-9]+)', form))
            tconstants.extend(re.findall('(1\.0)', form))
        
        allConstants = map(float, tconstants)
        allConstants = list(set(allConstants)) #remove duplicates
        if rawData.logic == 'bl' and allConstants != []:
                print "WARNING!! you have specified rational elements in a BL logic. The results may be uncontrolled"

        #now we can rewrite the constant symbols and associate a dictionary for each component
        
        #print allConstants

        constantsMap = dict()
        i = 0

        #print "data:allconstants: "
        #print allConstants
        
        newDataOperation  = rawData.dataOperation
        for ct in allConstants:
            dataOperation = newDataOperation
            newDataOperation = []
            constantsMap["c"+str(i)] = ct
#            print "constant: "
#            print ct
#            print "i: " + str(i)
#            
            for form in dataOperation:
                #print "form: "  + form
                form1 = form.replace(str(ct), "c"+str(i))
                newDataOperation.append(form1)
            i = i + 1
    
#        
#        for inp in rawData.dataOperation:
#            form = inp
#            for ct in allConstants:
#                form1 = form.replace(str(ct), "c"+str(i))
#                form = form1
#                #nf1 = inp.replace(str(ct), "c"+str(i))
#                
#                #nf2 = nf1.replace(" 1", " T")
#                #nf3 = nf2.replace(" 0", " F")
#                #print "nf1 :"+nf1
#                #print "nf2 :"+nf2
#                #print "nf3 :"+nf3
#                #dataOperation.append(nf1)
#                
#                i = i+1
#            dataOperation.append(form)

            
        self.dataOperation = newDataOperation
        self.constantsMap = constantsMap
        
        self.constantsInProducts = dict()
        for i in self.ordinalSum:
            if i[1] == "p":
                constantsini = [c for c in constantsMap if self._vinc(i, constantsMap[c])]
                if constantsini != []:
                    self.constantsInProducts[self.ordinalSum.index(i)] = [[c, constantsMap[c]] for c in constantsMap if self._vinc(i, constantsMap[c])]

        return
        
    #split the equation (R f1 f2) in a list [R, [f1, f2]], and stores this in self.spEquations
    def _splitEquation(self, eq):
        i = 1
        while eq[i] != " ":
            i = i +1
        d = i+1
        
        if eq[d] == "(":
                openPars = 1
                closedPars = 0
                d = d+1
                while closedPars != openPars:
                    if eq[d] == "(":
                        openPars = openPars + 1
                    if eq[d] == ")":
                        closedPars = closedPars +1
                    d = d+1
                        #d = d+1
        else:
            #phi is atomic
            while eq[d] != " " and eq[d] != ")":
                    d = d+1
        #store the "splitted" equations for future use
        self.spEquations.append([eq[1:i], [eq[i+1:d], eq[d+1:len(eq)-1]]])
        return
#        
#        
#        for i in range(len(self.ordinalSum)):
#            constantsIn.append([])
#            cborrar = []
#            for cname in constantsMap:
#                if self._getComponentIndex(constantsMap[cname]) == i:
#                    constantsIn[i].append([cname, constantsMap[cname]])
#                    cborrar.append(cname)
#                    #del constantsMap[cname]
#            for cname in cborrar:
#                del constantsMap[cname]
#        self.constantsIn = constantsIn
#
#        return
    
  ####################################################################
   ##FUNCTIONS CONCERNING THE CONSTRUCTION OF THE SKELETON FRAME
    def text_getArity(self, function):
        if (function == "delta" or function == "neg" or function == "box" or function == "diamond"):
            return 1
        else :
            return 2
            
    def text_isModal(self, function):
        if (function == "box" or function == "diamond"):
            return True
        else:
            return False
    
    #returns a pair with the operation name and, as second element, a list with the arguments -parsed recursively-
    #if it is a constant or a variable, it returns that
    def parseFormula(self, phi):
        #print "phi:" + phi
        
        if phi[0] != "(":
            #print "returned :" + phi
            return phi
        else:
            i = 1
            while phi[i] != " ":
                i = i +1
            #arity of the operation
            ar = self.text_getArity(phi[1:i])
            #print "arity:" + str(ar)

            d = i+1
            if phi[d] == "(":
                openPars = 1
                closedPars = 0
                d = d+1
                while closedPars != openPars:
                    if phi[d] == "(":
                        openPars = openPars + 1
                    if phi[d] == ")":
                        closedPars = closedPars +1
                    d = d+1
                #d = d+1
            else:
            #phi is atomic
                while phi[d] != " " and phi[d] != ")":
                    d = d+1
            if ar == 1:
                #print "called1 :" + phi
                return [phi[1:i], [self.parseFormula(phi[i+1:d])]]
            else :
                #print "called2 :" + phi
                return [phi[1:i], [self.parseFormula(phi[i+1:d]), self.parseFormula(phi[d+1:len(phi)-1])]]
    
    #parses the splitted equation [R, [f1, f2]]  and returns
    #a pair with the operation and the two argument formulas parsed             
    def parseEquation(self, eq):
        return [eq[0], [self.parseFormula(eq[1][0]), self.parseFormula(eq[1][1])]]
                
    #returns a list with the subformulas of formula up to propositional level
    def pSubformulas(self, formula):
        if type(formula) == str:
            return [formula]
        else:
            if not(self.text_isModal(formula[0])) :
                subs = []
                for g in formula[1]:
                    subs.extend(self.pSubformulas(g))
                return subs
            else:
                return [formula]
            
    def formula_isModal(self, formula):
        if type(formula) == str:
            return False
        else:
            return self.text_isModal(formula[0])
            
    #returns the modal depth (maximum level of nested modalities) of the formula
    def modalDepth(self, formula):
        if not(self.text_isModal(formula[0])):
            return max([self.modalDepth(g) for g in formula[1]])
        else:
            return 1 + self.modalDepth(formula[1][0])
         
    def formToText(self, formula):
        if type(formula) == str:
            return formula
        else:
            ar = self.text_getArity(formula[0])
            if ar == 1:
                return "("+formula[0]+" "+self.formToText(formula[1][0])+")"
            else:
                return "("+formula[0]+" "+self.formToText(formula[1][0])+" "+self.formToText(formula[1][1])+")"
    
    #from a modal formula (in text format), returns the non-modal part            
    def nonModal(self, mform):
        i = 1
        while mform[i] != " ":
            i = i +1
        return mform[i+1:-1]    

                
    #builds recursively the skeleton of the tree
    #mformula is expected to be a modal formula, 
    #Rprev is expected to be the tree over which include the new part, and originalWorld is the index of the root    
    def skForm(self, mformula, Rprev, originalWorld):
        #print "mformula: "
        #print mformula
        #print "originalWorld: "
        #print originalWorld
        
        nR = Rprev
        for i in range(len(Rprev)):
            if i != originalWorld:
                nR[i][2].append(0)
            else:
                nR[i][2].append(1)
        relations = []
        for i in range(len(Rprev)):
            relations.append(0)
        #one more for the new world
        relations.append(0)
        nR.append([self.formToText(mformula), originalWorld, relations])
        
        #print "skeleton: "
        #print nR
        
        nonModalForm = mformula[1][0]
        #print "nonModal: "
        #print nonModalForm
        
        subforms = self.pSubformulas(nonModalForm)
        #print "subforms: "
        #print subforms
        
        msubforms = [g for g in subforms if self.formula_isModal(g)]
        #print "msubforms: "
        #print msubforms
        
        if msubforms == []:
            return nR
        
        current = len(nR)-1
        for fm in msubforms:
            #print current
            #print fm
            
            nnR = self.skForm(fm, nR, current)
            nR = nnR

        return nR           
    
    #merges the skeletons from the list on the same root (without an assigned formula). 
    #The result has as dimmension the addition of all of them      
    def mergeSk(self, skList):
        tSize = 0
        R = [["",-1, [0]]]
        tSize = 1
        for sk in skList:
            tSize = tSize + len(sk[0][2])-1
            R[0][2].extend(sk[0][2][1:])
        k = 1
        lenPrevs = 0
        
        #print "tSize:" + str(tSize)
        #print "skList:"
        #print skList
        for i in range(len(skList)):
            if i > 0:
                lenPrevs = lenPrevs + len(skList[i-1])-1
            #print skList[i]
            for j in range(len(skList[i])-1):
                #print "j: "+str(j)
                R.append([skList[i][j+1][0], 0, []])
                
                if skList[i][j+1][1] == 0:
                    R[k][1] = 0
                else:
                    R[k][1] = skList[i][j+1][1] + lenPrevs
                R[k][2] = [0 for index in range(tSize)]
                R[k][2][lenPrevs:lenPrevs + len(skList[i][j+1][2])] = skList[i][j+1][2]
                #print "R"
                #print R
                k = k + 1
        return R
                               
    
           
    #the structure skeleton comprehends more than the basic structure over which the solver will work. 
    #It encodes an enhancement that reduces the number of tests (a label for each world with a modal formula), 
    #and the father of each world, prior to the list of accessibility
    # each world, except for the root, is associated with one modal formula. The value on the father of the world of that formula
    # is the value on that world. 
    def _buildSkeleton(self):
        if self.modal == 0:
            self.skeleton = [["",-1,[0]]]
            self.nworlds = 1
            return
        if self.operation == "th" or self.operation == "d":
            #formulas
            fparsed = [[phi, self.parseFormula(phi)] for phi in self.dataOperation]
            #print fparsed
            #modal subformulas
            sbfs = []
            for form in fparsed:
                psubs = self.pSubformulas(form[1])
                msubs = [mf for mf in psubs if self.formula_isModal(mf)]
                sbfs.extend(msubs)
            #if sbfs == []:
            #    self.skeleton =  [[[""],[]]]
            #else :
            #    return mergeSk([skForm(phi, [[[""],[]]], 0) for phi in sbfs])
        #for modal logics, we will undestand satisfiability as local satisfiability, and this will 
        #happen only in an structure with the skeleton determined by the fromulas from the equations
        if self.operation == "s":
            eqparsed = [self.parseEquation(eq) for eq in self.spEquations]
            fparsed = []
            for eq in eqparsed:
                fparsed.extend(eq[1])
            
            #print "fparsed: "
            #print fparsed
            #modal subformulas
            sbfs = []

            for phi in fparsed:
                sbfs.extend([mf for mf in self.pSubformulas(phi) if self.formula_isModal(mf)])
        #print sbfs
        if sbfs == []:
            self.skeleton =  [["",-1, [0]]]
        else :
            self.skeleton = self.mergeSk([self.skForm(phi, [["",-1,[0]]], 0) for phi in sbfs])
        self.nworlds = len(self.skeleton)

        #print "skeleton"
        #print self.skeleton

        return

            